//
// pysearch.cc
//
// Python module to search the htdig databases.
// Expects the databases are generated using htdig, htmerge, and htfuzzy
// search method returns a tuple of (hits, vars) to caller. 
//
// hits is a list of dictionaries which contains the data for matching pages.
//   SIZE
//   HOPCOUNT
//   DESCRIPTIONS
//   SCORE
//   CURRENT
//   DESCRIPTION
//   MODIFIED
//   EXCERPT
//   NUMSTARS
//   URL
//   TITLE
//   BACKLINKS
//   DOCID
//   SIZEK
//   PERCENT
//
// vars contains data that pertains to the whole set of data:
//   PLURAL_MATCHES
//   PAGE
//   PAGES
//   EXCLUDE
//   KEYWORDS
//   MATCHES_PER_PAGE
//   MATCH_MESSAGE
//   VERSION
//   RESTRICT
//   LASTDISPLAYED
//   MATCHES
//   MAX_STARS
//   FIRSTDISPLAYED
//
// This add on to htdig uses a lot of functionality that was 
// already built into htdig, but had to be pulled out in a kind of
// odd way, since it wasn't very modular.  The file HtSearchFuncs.cc
// is a set of functions that were pulled out of the file htsearch.cc.
//
// This file was also based on the htsearch.cc file, but needed lots
// of modification to accept python data structures as argument, and
// instead of returning an HTML page, it just returns a set of data
// that the calling program can do with as it pleases.
//
// The PyDisplay class is a rip off of the Display class in htsearch.
// It seems like the Display class is actually overloaded a bit, 
// doing result weightings, as well as formatting the results as
// HTML to return to the browser.  What I really wanted was just the
// weighting part and the excerpt generation part, but by inheriting
// from the Display class, I got a lot of stuff that's not being used.
// 
//
//
#if RELEASE
static char RCSid[] = "$Id: pysearch.cc,v 1.2 2001/12/08 23:20:19 erickson Exp erickson $";
#endif


#include "Python.h"
#include "parser.h"
#include "PyDisplay.h"
#include <signal.h>
#include "HtURLCodec.h"
#include "HtWordType.h"
#include "HtSearchFuncs.h"

// If we have this, we probably want it.
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

typedef void (*SIGNAL_HANDLER) (...);


int			debug = 3;

//extern ResultList* htsearch(char *wordfile, List &searchWords, Parser *parser);


typedef struct input_list {
  char* restrict;
  char* exclude;
  char* config;
  char* method;
  char* keywords;
  char* sort;
  char* words;
  int matchesperpage;
  int page;
};


static PyObject * search (PyObject *self, PyObject *args, PyObject *kwargs)
{
    int			c;
    extern char		*optarg;
    int		        override_config=0;
    List		searchWords;
    String		configFile = DEFAULT_CONFIG_FILE;
    int			pageNumber = 1;
    StringMatch		limit_to;
    StringMatch		exclude_these;
    String		logicalWords;
    String              origPattern;
    String              logicalPattern;
    StringMatch		searchWordsPattern;
    StringList		requiredWords;
    int                 i;

    input_list inputs;

    static char * empty="";
    inputs.restrict = empty;
    inputs.exclude = empty;
    inputs.config = empty;
    inputs.method = empty;
    inputs.keywords = empty;
    inputs.sort = empty;
    inputs.words = empty;
    inputs.matchesperpage = 0;
    inputs.page = 0;

    static char * argnames[] =
    {
      "restrict", 
      "exclude",
      "config",
      "method",
      "matchesperpage",
      "page",
      "keywords",
      "sort",
      "words"
    };


    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssssiisss", argnames,
			   &inputs.restrict,
			   &inputs.exclude,
			   &inputs.config,
			   &inputs.method,
			   &inputs.matchesperpage,
			   &inputs.page,
			   &inputs.keywords,
			   &inputs.sort,
			   &inputs.words))
    {
	PyErr_SetString(PyExc_RuntimeError, "Error Parsing Arguments");
        return NULL;
    }
    
    //
    // The total search can NEVER take more than 5 minutes.
    //
//    alarm(5 * 60);

    //
    // Compile the URL limit pattern.
    //
    if (inputs.restrict && strlen(inputs.restrict) > 0)
    {
	char *sep = inputs.restrict;
	while ((sep = strchr(sep, '\001')) != NULL)
	  *sep++ = '|';
	limit_to.Pattern(inputs.restrict);
    }
    if (inputs.exclude && strlen(inputs.exclude) > 0)
    {
       char *sep = inputs.exclude;
       while ((sep = strchr(sep, '\001')) != NULL)
       	 *sep++ = '|';
       exclude_these.Pattern(inputs.exclude);
    }

    //
    // Setup the configuration database.  First we read the compiled defaults.
    // Then we override those with defaults read in from the configuration
    // file, and finally we override some attributes with information we
    // got from the HTML form.
    //
    config.Defaults(&defaults[0]);

    if (inputs.config && strlen(inputs.config) > 0)
    {
	configFile = inputs.config;
    }

    // To allow . in filename while still being 'secure',
    // e.g. htdig-f.q.d.n.conf
/*    if (!override_config && input.exists("config") 
	&& (strstr(input["config"], "./") == NULL))
    {
	char	*configDir = getenv("CONFIG_DIR");
	if (configDir)
	{
	    configFile = configDir;
	}
	else
	{
	    configFile = CONFIG_DIR;
	}
	if (strlen(input["config"]) == 0)
	  configFile = DEFAULT_CONFIG_FILE;
	else
	  configFile << '/' << input["config"] << ".conf";
    }
*/
    if (access(configFile, R_OK) < 0)
    {
	PyErr_SetString(PyExc_ValueError, 
			form("Unable to read configuration file '%s'",
			 configFile.get()));
        return NULL;
    }
    config.Read(configFile);

    if (inputs.method)
    {
	String tmp(inputs.method);
	tmp.lowercase();
	config.Add("match_method", tmp);
    }

    if (inputs.matchesperpage > 0)
    {
	// minimum check for a valid int value of "matchesperpage" cgi variable
	//if (atoi(input["matchesperpage"]) > 0)
	char mpp[10];
	sprintf(mpp, "%4.4d", inputs.matchesperpage);
	config.Add("matches_per_page", mpp);
    }

    if (inputs.page)
	pageNumber = inputs.page;
    if (inputs.config)
	config.Add("config", inputs.config);
    if (inputs.restrict)
	config.Add("restrict", inputs.restrict);
    if (inputs.exclude)
	config.Add("exclude", inputs.exclude);
    if (inputs.keywords)
	config.Add("keywords", inputs.keywords);

    requiredWords.Create(config["keywords"], " \t\r\n\001");

    if (inputs.sort)
	config.Add("sort", inputs.sort);

    StringList form_vars(config["allow_in_form"], " \t\r\n");
//    for (i= 0; i < form_vars.Count(); i++)
//   {
//     if (input.exists(form_vars[i]))
//	config.Add(form_vars[i], input[form_vars[i]]);
//   }
 
    // Ctype-like functions for what constitutes a word.
    HtWordType::Initialize(config);

    //
    // Check url_part_aliases and common_url_parts for
    // errors.
    String url_part_errors = HtURLCodec::instance()->ErrMsg();

    if (url_part_errors.length() != 0)
    {
	PyErr_SetString(PyExc_RuntimeError, 
			form("Invalid url_part_aliases or common_url_parts: %s",
                          url_part_errors.get()));
        return NULL;
    }

    Parser	*parser = new Parser();
	
    //
    // Parse the words to search for from the argument list.
    // This will produce a list of WeightWord objects.
    //
    String	 originalWords = inputs.words;
    originalWords.chop(" \t\r\n");
    setupWords(originalWords, searchWords,
	       strcmp(config["match_method"], "boolean") == 0,
	       parser, origPattern);

    if (searchWords.Count() < 1)
    {
	PyErr_SetString(PyExc_RuntimeError, 
	   "Invalid Boolean expression");
        return NULL;
    }

    //
    // Convert the list of WeightWord objects to a pattern string
    // that we can compile.
    //
    createLogicalWords(searchWords, logicalWords, logicalPattern);

    // 
    // Assemble the full pattern for excerpt matching and highlighting
    //
    origPattern += logicalPattern;
    searchWordsPattern.IgnoreCase();
    searchWordsPattern.IgnorePunct();
    searchWordsPattern.Pattern(logicalPattern);	// this should now be enough
    //searchWordsPattern.Pattern(origPattern);
    //if (debug > 2)
    //  cout << "Excerpt pattern: " << origPattern << "\n";

    //
    // If required keywords were given in the search form, we will
    // modify the current searchWords list to include the required
    // words.
    //
    if (requiredWords.Count() > 0)
    {
	addRequiredWords(searchWords, requiredWords);
    }
    
    //
    // Perform the actual search.  The function htsearch() is used for this.
    // The Dictionary it returns is then passed on to the PyDisplay object to
    // store the results in a python list.
    //
    String	word_db = config["word_db"];
    if (access(word_db, R_OK) < 0)
    {
	PyErr_SetString(PyExc_RuntimeError, 
            form("Unable to read word database file '%s'\nDid you run htmerge?",
			 word_db.get()));
        return NULL;
    }
    ResultList	*results = htsearch(word_db, searchWords, parser);

    String	index = config["doc_index"];
    if (access(index, R_OK) < 0)
    {
	PyErr_SetString(PyExc_RuntimeError, 
	   form("Unable to read document index file '%s'\nDid you run htmerge?",
			 index.get()));
        return NULL;
    }
    String	doc_db = config["doc_db"];
    if (access(doc_db, R_OK) < 0)
    {
	PyErr_SetString(PyExc_RuntimeError, 
            form("Unable to read document database file '%s'\nDid you run htmerge?",
	            doc_db.get()));
        return NULL;
    }

    
    PyDisplay	display(index, doc_db);
    PyObject* pVarsList=PyDict_New();
    PyObject* pHitList=PyList_New(0);

    display.setOriginalWords(originalWords);
    display.setResults(results);
    display.setSearchWords(&searchWords);
    display.setLimit(&limit_to);
    display.setExclude(&exclude_these);
    display.setAllWordsPattern(&searchWordsPattern);

    display.setLogicalWords(logicalWords);
    if (parser->hadError())
    {
	display.displaySyntaxError(parser->getErrorMessage());
    }
    else
    {
        display.display(pVarsList, pHitList, pageNumber);
    }


    
    delete results;
    delete parser;
    PyObject* ret=PyTuple_New(2);
    PyTuple_SetItem(ret, 0, pHitList);
    PyTuple_SetItem(ret, 1, pVarsList);
    return ret;
}

//
//  Let python know what methods are available.
//
static struct PyMethodDef pysearch_methods[] = {
	{"search", (PyCFunction)search, METH_VARARGS|METH_KEYWORDS},
	{NULL, NULL}
};

// So that the name is not mangled.
extern "C" void initpysearch();

//
//  Initialize the python module.
//
void initpysearch()
{
    PyObject *m, *d;

    m=Py_InitModule("pysearch", pysearch_methods);

    d=PyModule_GetDict(m);

    if (PyErr_Occurred())
	    Py_FatalError("can't initialize module pysearch");
}



